__all__ = [
    'Cutter',
    'Dearrayer',
    'preprocess'
    'TileLabeler'
]

from ._cutter import Cutter
from ._dearrayer import Dearrayer
from ._labeler import TileLabeler
from . import preprocess

from ._version import __version__



import os
from os.path import dirname, join
import multiprocessing as mp
from typing import Union, List, Dict, Tuple
import logging

import pandas as pd
import numpy as np
import cv2
from PIL import Image, ImageDraw

import matplotlib.pyplot as plt
import ipywidgets as widgets
from ipywidgets import Layout, VBox, HBox
from IPython.display import display

__all__ = [
    'Explore',
    'plot_on_thumbnail',
    'plot_tiles',
    'plot_ranges',
]

# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def split_list(l: list, n: int = 1) -> list:
    """Split list into n parts"""
    length = len(l)
    split = []
    for i in range(n):
        split.append(l[i*length // n: (i+1)*length // n])
    return split


def get_load_image(nrows: int, ncols: int, px: int):
    """Generates a white image."""
    load_img = np.ones((nrows*px, ncols*px, 3))*255
    load_img = cv2.imencode('.png', load_img)[1].tobytes()
    return load_img


def get_tiles(
    df: pd.DataFrame,
    column: str = None,
    x_range: tuple = None,
    nrows: int = 2,
    ncols: int = 16,
    px: int = 64,
    to_bytes=True,
) -> bytes:
    """Return image grid of the tiles from selection."""
    # Collect random sample and tile paths.
    if column is None and x_range is None:
        paths = df.path
    else:
        paths = df.path[
            (df[column] > x_range[0]) &
            (df[column] < x_range[1])
        ]
    number_of_samples = min(ncols*nrows, len(paths.tolist()))
    paths = paths.sample(number_of_samples)
    # Load images
    tiles = []
    processes = min(ncols*nrows, os.cpu_count() - 1)
    with mp.Pool(processes) as p:
        for result in p.imap(cv2.imread, paths):
            tiles.append(result)
    # Resize.
    tiles = [cv2.resize(x, (px, px)) for x in tiles]
    # Add white images if they run out.
    while nrows*ncols > len(tiles):
        tiles.append(np.ones((px, px, 3)) * 255)
    # Combine.
    tiles = [np.hstack(r) for r in split_list(tiles, nrows)]
    tiles = np.vstack(tiles)
    if to_bytes:
        # To bytes.
        __, encoded = cv2.imencode('.png', tiles)
        return encoded.tobytes()
    else:
        tiles = cv2.cvtColor(tiles.astype(np.uint8), cv2.COLOR_BGR2RGB)
        return Image.fromarray(tiles)


def get_widgets(
    df: pd.DataFrame,
    group: str,
    nrows: int = 2,
    ncols: int = 16,
    px: int = 64,
):
    # Collect all columns.
    columns = {
        'sharpness': [x for x in df.columns if 'sharpness' in x],
        'data_loss': [x for x in df.columns if '_pixels' in x],
        'hue': [x for x in df.columns if 'hue_' in x],
        'sat': [x for x in df.columns if 'sat_' in x],
        'val': [x for x in df.columns if 'val_' in x],
        'red': [x for x in df.columns if 'red_' in x],
        'green': [x for x in df.columns if 'green_' in x],
        'blue': [x for x in df.columns if 'blue_' in x],
    }
    # Collect all options.
    col_options = {
        'sharpness': [(x.split('_')[-1], x) for x in columns['sharpness']],
        'data_loss': [(x.split('_')[0], x) for x in columns['data_loss']],
        'hue': [(x.split('_')[-1], x) for x in columns['hue']],
        'sat': [(x.split('_')[-1], x) for x in columns['sat']],
        'val': [(x.split('_')[-1], x) for x in columns['val']],
        'red': [(x.split('_')[-1], x) for x in columns['red']],
        'green': [(x.split('_')[-1], x) for x in columns['green']],
        'blue': [(x.split('_')[-1], x) for x in columns['blue']],
    }
    drop_options = [
        ('Hue', 'hue'),
        ('Saturation', 'sat'),
        ('Value/brightness', 'val'),
        ('Red', 'red'),
        ('Green', 'green'),
        ('Blue', 'blue'),
    ]
    # Prepare arguments for each plot group.
    args = {
        'sharpness': {
            'col_desc': 'Reduction:',
            'x_range': (
                df[columns['sharpness'][0]].min(),
                df[columns['sharpness'][0]].max()
            ),
            'slider': widgets.FloatRangeSlider,
            'format': '.1f',
            'xmin': df[columns['sharpness'][0]].min(),
            'xmax': df[columns['sharpness'][0]].max(),
            'step': 0.1,
        },
        'data_loss': {
            'col_desc': 'Data loss:',
            'x_range': (
                df[columns['data_loss'][0]].min(),
                df[columns['data_loss'][0]].max()
            ),
            'slider': widgets.FloatRangeSlider,
            'format': '.3f',
            'xmin': 0,
            'xmax': 1,
            'step': 0.001,
        },
        'channels': {
            'col_desc': 'Quantile:',
            'x_range': (0, 255),
            'slider': widgets.IntRangeSlider,
            'format': 'd',
            'xmin': 0,
            'xmax': 255,
            'step': 1,
        }
    }

    # Build plot widgets.
    if group == 'channels':
        col = columns['hue'][0]
        opt = col_options['hue']
    else:
        col = columns[group][0]
        opt = col_options[group]
    plot_wdgts = {
        'column': widgets.SelectionSlider(
            options=opt,
            description=args[group]['col_desc'],
            continuous_update=False,
        ),
        'bins': widgets.IntSlider(
            30, 10, 100,
            step=5, description='Bins:',
            continuous_update=False,
        ),
        'x_range': args[group]['slider'](
            value=args[group]['x_range'],
            min=args[group]['xmin'],
            max=args[group]['xmax'],
            step=args[group]['step'],
            description='Range:',
            continuous_update=False,
            readout_format=args[group]['format'],
        ),
        'log_y': widgets.ToggleButton(
            value=False,
            description='log y',
            button_style='',
            icon=''
        )
    }
    if group == 'channels':
        plot_wdgts['dropdown'] = widgets.Dropdown(
            options=drop_options,
            value=drop_options[0][1],
            description='Plot:',
        )
    else:
        plot_wdgts['dropdown'] = None

    # Build tiles_widgets
    tile_wdgts = {
        'nrows': widgets.Dropdown(
            options=[2, 4, 8],
            value=2,
            description='Rows:',
        ),
        'ncols': widgets.Dropdown(
            options=[4, 8, 16, 32, 64],
            value=16,
            description='Columns:',
        ),
        'px': widgets.Dropdown(
            options=[32, 64, 128, 256],
            value=64,
            description='Pixel width:',
        )
    }
    # Load first tiles.
    x_range = (args[group]['xmin'], args[group]['xmax'])
    first_tiles = get_tiles(df, col, x_range, nrows, ncols, px)
    # Create image widget
    tiles = widgets.Image(
        value=first_tiles,
        layout=widgets.Layout(width='100%')
    )
    tiles.__dict__['column'] = col
    tiles.__dict__['x_range'] = args[group]['x_range']
    tiles.__dict__['nrows'] = nrows
    tiles.__dict__['ncols'] = ncols
    tiles.__dict__['px'] = px
    # Link widgets to tile images.

    def update_tiles(change):
        for key, val in tiles.__dict__.items():
            if val == change['old']:
                tiles.__dict__[key] = change['new']

    plot_wdgts['column'].observe(update_tiles, names='value')
    plot_wdgts['x_range'].observe(update_tiles, names='value')
    tile_wdgts['nrows'].observe(update_tiles, names='value')
    tile_wdgts['ncols'].observe(update_tiles, names='value')
    tile_wdgts['px'].observe(update_tiles, names='value')

    # Button to plot new tiles.
    button_tiles = widgets.Button(
        description="Refresh tiles!",
    )

    def on_button_tiles_clicked(b):
        tiles.value = get_load_image(
            nrows=tiles.__dict__['nrows'],
            ncols=tiles.__dict__['ncols'],
            px=tiles.__dict__['px']
        )
        tiles.value = get_tiles(
            df,
            tiles.__dict__['column'],
            tiles.__dict__['x_range'],
            nrows=tiles.__dict__['nrows'],
            ncols=tiles.__dict__['ncols'],
            px=tiles.__dict__['px']
        )
    button_tiles.on_click(on_button_tiles_clicked)

    # Selection text.
    selection = widgets.Textarea(
        value='',
        placeholder="Copypastable selection will appear here!",
        description='Selection:',
        layout=widgets.Layout(width='100%', height='100px')
    )
    selection.__dict__['selections'] = []

    # Button to add stuff to selection.
    button_selection = widgets.Button(
        description="Add to selection!",
    )

    def on_button_selection_clicked(b):
        # Get widget values.
        column = plot_wdgts['column'].value
        x_range = plot_wdgts['x_range'].value
        # Add selection text to selections.
        if len(selection.__dict__['selections']) > 0:
            selection.__dict__['selections'][-1] += ' |'
        selection.__dict__['selections'].append(
            f"    "
            f"((df['{column}'] >= {x_range[0]}) & "
            f"(df['{column}'] <= {x_range[1]}))"
        )
        # Prepare suffix, prefix and indents.
        prefix = ['bad_tiles = df[(']
        suffix = [')]']
        # Build text.
        string = '\n'.join(prefix + selection.__dict__['selections'] + suffix)
        selection.value = string

    button_selection.on_click(on_button_selection_clicked)

    button_clear = widgets.Button(
        description="Clear selection!",
    )

    def on_button_clear_clicked(b):
        # Clear selections.
        selection.__dict__['selections'] = []
        selection.value = ''
    button_clear.on_click(on_button_clear_clicked)

    # Link drop menu.
    def update_column_options(change):
        plot_wdgts['column'].options = col_options[change['new']]
        plot_wdgts['column'].value = columns[change['new']][0]

    if plot_wdgts['dropdown'] is not None:
        plot_wdgts['dropdown'].observe(update_column_options, names='value')

    return (
        plot_wdgts,
        tiles, button_tiles, tile_wdgts,
        selection, button_selection, button_clear
    )


def get_ui(
    plot_wdgts: dict,
    tile_wdgts: dict,
    button_tiles: widgets.Widget,
    button_selection: widgets.Widget,
    button_clear: widgets.Widget,
):
    # Change layout:
    plot_wdgts['column'].layout = widgets.Layout(width='100%')
    plot_wdgts['bins'].layout = widgets.Layout(width='100%')
    plot_wdgts['x_range'].layout = widgets.Layout(width='100%')
    plot_wdgts['log_y'].layout = widgets.Layout(width='10%')
    tile_wdgts['nrows'].layout = widgets.Layout(width='100%')
    tile_wdgts['ncols'].layout = widgets.Layout(width='100%')
    tile_wdgts['px'].layout = widgets.Layout(width='100%')
    button_tiles.layout = widgets.Layout(width='100%')
    button_selection.layout = widgets.Layout(width='20%')
    button_clear.layout = widgets.Layout(width='20%')

    if plot_wdgts['dropdown'] is None:
        top = HBox([
            plot_wdgts['column'],
            plot_wdgts['bins']
        ])
        del plot_wdgts['dropdown']
    else:
        plot_wdgts['dropdown'].layout = widgets.Layout(width='50%')
        top = HBox([
            plot_wdgts['dropdown'],
            plot_wdgts['column'],
            plot_wdgts['bins']
        ])
    middle = HBox([
        plot_wdgts['x_range'],
        plot_wdgts['log_y'],
        button_selection,
        button_clear,
    ])
    bottom_ui = HBox([
        button_tiles,
        tile_wdgts['nrows'],
        tile_wdgts['ncols'],
        tile_wdgts['px'],
    ])
    top_ui = VBox([top, middle])
    return top_ui, bottom_ui


def get_plot_func(df, dropdown=False):
    if dropdown:
        def plot_func(column, bins, x_range, log_y, dropdown):
            x = df[column][
                (df[column] >= x_range[0]) &
                (df[column] <= x_range[1])
            ]
            plt.figure(figsize=(15, 5))
            plt.hist(x, bins=bins)
            plt.xlim(x_range)
            if log_y and len(x) > 0:
                plt.yscale('log')
            plt.show()
    else:
        def plot_func(column, bins, x_range, log_y):
            x = df[column][
                (df[column] >= x_range[0]) &
                (df[column] <= x_range[1])
            ]
            plt.figure(figsize=(15, 5))
            plt.hist(x, bins=bins)
            plt.xlim(x_range)
            if log_y and len(x) > 0:
                plt.yscale('log')
            plt.show()
    return plot_func


def Explore(
    metadata: pd.DataFrame,
    sharpness: bool = False,
    data_loss: bool = False,
    channels: bool = False,
):
    """Explore metadata to discover outliers.

    This function can be used for easy exploration of the metadata 
    generated during cutting of the tiles. By exploring the histograms
    it is easy to detect outlying values in certain preprocessing 
    metrics.

    Args:
        metadata (pd.DataFrame): Metadata generated during cutting.
        sharpness (bool, optional): Explore sharpness. 
            Defaults to False.
        data_loss (bool, optional): Explore data loss. 
            Defaults to False.
        channels (bool, optional): Explore different channel quantiles.
            Defaults to False.

    Raises:
        ValueError: ``sharpness``, ``data_loss`` and ``channels`` are 
            all set to False.
        ValueError: More than one from ``sharpness``, ``data_loss`` and
        ``channels`` is set to ``True``.
    """
    # Check options.
    options = ['sharpness', 'data_loss', 'channels']
    if sum([sharpness, data_loss, channels]) == 0:
        raise ValueError(
            f'Please set one of the options {options} to True.')
    if sum([sharpness, data_loss, channels]) > 1:
        raise ValueError(
            f'Please set only one of the options {options} to True.')
    if sharpness:
        what_to_plot = 'sharpness'
        dropdown = False
    if data_loss:
        what_to_plot = 'data_loss'
        dropdown = False
    if channels:
        what_to_plot = 'channels'
        dropdown = True

    plot_w, t, b_t, tile_w, selection_w, b_s, b_c = get_widgets(
        metadata, what_to_plot)
    top_ui, bottom_ui = get_ui(plot_w, tile_w, b_t, b_s, b_c)
    plot_func = get_plot_func(metadata, dropdown)
    out = widgets.interactive_output(plot_func, plot_w)
    display(top_ui, out, bottom_ui, t, selection_w)


def plot_on_thumbnail(
    df: pd.DataFrame,
    max_thumbnails: int = 5,
    min_tiles: int = None,
) -> dict:
    """Plot tiles in the metadata dataframe onto thumbnails.

    Args:
        df (pd.DataFrame): Pandas dataframe with metadata.
        max_thumbnails (int, optional): Maximum number of thumbnails to
            find. Defaults to 5.
        min_tiles (int, optional): Minimum number of tiles 
            necessary to plot on thumbnail. Defaults to None.

    Raises:
        TypeError: Dataframe is the wrong type.

    Returns:
        dict: Dictionary with annotated slide thumbnails.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError('Expected {} not {}'.format(
            pd.DataFrame, type(df)
        ))
    # See if there are multiple slides.
    slides = df.slide_name.unique()
    if min_tiles is not None:
        # Discard slides with too few occurrences.
        new_slides = []
        for slide in slides:
            occurrences = len(df[df.slide_name == slide])
            if occurrences >= min_tiles:
                new_slides.append(slide)
        if len(new_slides) == 0:
            logger.warning(
                f'No slides found with minimum of {min_tiles} tiles!')
        slides = [x for x in slides if len(
            df[df.slide_name == x]) >= min_tiles]
    if len(slides) > max_thumbnails:
        slides = np.random.choice(slides, max_thumbnails, replace=False)
    # Load thumbnails.
    thumbnails = {}
    for slide in slides:
        tile_path = df[df.slide_name == slide].path.tolist()[0]
        data_path = os.path.dirname(tile_path.split('tiles')[0])
        # Find thumbnail and downsample.
        for x in os.scandir(data_path):
            if 'thumbnail' in x.name:
                downsample = x.name.split('_')[-1].split('.')[0]
                if downsample.isnumeric():
                    thumbnails[slide] = (Image.open(x.path), int(downsample))

    # Generate annotations.
    annotated_thumbnails = {}
    for slide in slides:
        # Prepare thumbnail.
        thumbnail, downsample = thumbnails[slide]
        annotated_thumbnail = thumbnail.copy()
        annotated = ImageDraw.Draw(annotated_thumbnail)
        # Get tile coords.
        tmp = df[df.slide_name == slide]
        width = tmp['width']
        x = tmp['x']
        y = tmp['y']
        if 'spot_path' in tmp.columns:
            # Try to load spot metadata.
            data_path = os.path.dirname(tile_path.split('tiles')[0])
            spot_meta_path = os.path.join(data_path, 'spot_metadata.csv')
            try:
                spot_meta = pd.read_csv(spot_meta_path)
            except Exception as e:
                raise IOError(f'Could not load {spot_meta_path}.')
            # Merge with tmp so we can add spot coords to tile coords.
            tmp = pd.merge(
                tmp, spot_meta, left_on='spot_path', right_on='path')
            x += tmp['x_y']
            y += tmp['y_y']
        coords = list(zip(x, y, width))
        # Draw.
        for x, y, width in coords:
            w_d = width/downsample
            x_d = x/downsample
            y_d = y/downsample
            annotated.rectangle(
                [x_d, y_d, x_d+w_d, y_d+w_d],
                outline='blue', width=2
            )
        annotated_thumbnails[slide] = annotated_thumbnail
    return annotated_thumbnails


def plot_tiles(df: pd.DataFrame, nrows: int = 4, ncols: int = 16, px: int = 64) -> Image.Image:
    """Return a random collection of tiles from given metadata dataframe.

    Args:
        df (pd.DataFrame): Dataframe with metadata
        nrows (int, optional): Number of rows. Defaults to 4.
        ncols (int, optional): Number of columns. Defaults to 16.
        px (int, optional): Width of individual tiles. Defaults to 64.

    Raises:
        TypeError: Dataframe is the wrong type.

    Returns:
        Image.Image: Collection of random tiles.
    """

    if not isinstance(df, pd.DataFrame):
        raise TypeError('Expected {} not {}'.format(
            pd.DataFrame, type(df)
        ))
    if len(df) == 0:
        logger.warning('No images.')
        return
    df = df.sample(n=min(len(df), nrows*ncols))
    tiles = get_tiles(df, nrows=nrows, ncols=ncols, px=px, to_bytes=False)
    return tiles


def plot_ranges(
    df: pd.DataFrame,
    column: str,
    ranges: List[Tuple[float]],
    nrows: int = 4,
    ncols: int = 8,
    px: int = 64,
) -> dict:
    """Plot random images from ranges from the given column,

    Args:
        df (pd.DataFrame): Dataframe with metadata.
        column (str): Desired column from the dataframe.
        ranges (List[Tuple[float]]): Ranges to be plotted  in format
            [(low,high), ...]
        nrows (int, optional): Number of rows. Defaults to 4.
        ncols (int, optional): Number of columns. Defaults to 16.
        px (int, optional): Width of individual tiles. Defaults to 64.

    Raises:
        TypeError: Dataframe is the wrong type.
        ValueError: Column not in dataframe.
        ValueError: Ranges in the wrong format.

    Returns:
        dict: [description]
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError('Expected {} not {}'.format(
            pd.DataFrame, type(df)
        ))
    if column not in df.columns:
        raise ValueError(f'Column {column} not in the given dataframe.')
    range_tiles = {}
    for low, high in ranges:
        if low > high:
            raise ValueError(
                f'{low} > {high}. Please give list of ranges in format '
                '[(low,high), (low,high), ...]'
            )
        tiles = get_tiles(df, column=column, x_range=(low, high),
                          nrows=nrows, ncols=ncols, px=px, to_bytes=False)
        range_tiles[f'{low}-{high}'] = tiles
    return range_tiles



import os
import itertools
import multiprocessing as mp
from functools import partial
from typing import Tuple, List, Callable, Union
import logging

import cv2
from openslide import OpenSlide
import numpy as np
from PIL import Image
from tqdm import tqdm
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import silhouette_score

from .preprocess.functional import tissue_mask, PIL_to_array
from ._czi_reader import OpenSlideCzi

########################
### Common functions ###
########################

# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_downsamples(slide_path: str) -> dict:
    reader = OpenSlide(slide_path)
    downsamples = [round(x) for x in reader.level_downsamples]
    dims = [x for x in reader.level_dimensions]
    return dict(zip(downsamples, dims))


def get_thumbnail(
    slide_path: str,
    downsample: int = None,
    create_thumbnail: bool = False
) -> Image.Image:
    """Return thumbnail by using openslide or by creating a new one."""
    if slide_path.endswith('czi'):
        if create_thumbnail:
            return generate_thumbnail(slide_path, downsample)
        else:
            return None
    else:
        reader = OpenSlide(slide_path)
        level_downsamples = [round(x) for x in reader.level_downsamples]
        if downsample is None:
            downsample = max(level_downsamples)
        if downsample in level_downsamples:
            level = level_downsamples.index(downsample)
            dims = reader.level_dimensions[level]
            thumbnail = reader.get_thumbnail(dims)
        elif create_thumbnail:
            thumbnail = generate_thumbnail(slide_path, downsample)
        else:
            thumbnail = None
        return thumbnail


def generate_thumbnail(
    slide_path: str,
    downsample: int,
    width: int = 4096
) -> Image.Image:
    """Generate thumbnail for a slide."""
    if slide_path.endswith('czi'):
        reader = OpenSlideCzi(slide_path)
        thumbnail = reader.get_thumbnail(downsample, 2048, fast=True)
        return thumbnail
    else:
        reader = OpenSlide(slide_path)
        dims = reader.dimensions
        blocks = (
            int(dims[0]/width) + 1,
            int(dims[1]/width) + 1
        )
        x = (i*width for i in range(blocks[0]))
        y = (i*width for i in range(blocks[1]))
        coords = list(enumerate(itertools.product(x, y)))
        # Multiprocessing to make things speedier.
        with mp.Pool(processes=os.cpu_count()) as p:
            func = partial(load_tile, slide_path, width, downsample)
            tiles = []
            for result in tqdm(
                p.imap(func, coords),
                total=len(coords),
                desc='Generating thumbnail',
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                tiles.append(result)
        # Collect each column seperately and mash them together.
        all_columns = []
        col = []
        for i, tile in tiles:
            if i % blocks[1] == 0 and i != 0:
                all_columns.append(np.vstack(col))
                col = []
            col.append(np.array(tile))
        thumbnail = np.hstack(all_columns)
        # Turn into into pillow image.
        thumbnail = Image.fromarray(thumbnail.astype(np.uint8))
        return thumbnail


def load_tile(
        slide_path: str,
        width: int,
        downscale: int,
        coords: Tuple[int, Tuple[int, int]]):
    # Load slide from global.
    reader = OpenSlide(slide_path)
    i, (x, y) = coords
    out_shape = (int(width/downscale), int(width/downscale))
    try:
        tile = reader.read_region((x, y), 0, (width, width)).convert('RGB')
        tile = cv2.resize(np.array(tile), out_shape, cv2.INTER_LANCZOS4)
    except:
        tile = np.zeros(out_shape + (3,)).astype(np.uint8)
    return i, tile


def resize(image: Union[np.ndarray, Image.Image], max_pixels: int = 1_000_000):
    """Donwsaple image until it has less than max_pixels pixels."""
    if isinstance(image, Image.Image):
        image = PIL_to_array(image)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    dimensions = image.shape[:2]
    width, height = dimensions
    factor = 0
    while width*height > max_pixels:
        factor += 1
        width = int(dimensions[0]/2**factor)
        height = int(dimensions[1]/2**factor)
    image = Image.fromarray(image)
    return image.resize((height, width))


def try_thresholds(
    thumbnail: Image.Image,
    thresholds: List[int],
    max_pixels=1_000_000
) -> Image.Image:
    """Returns a summary image of different thresholds."""
    thumbnail = resize(thumbnail)
    gray = cv2.cvtColor(np.array(thumbnail), cv2.COLOR_RGB2GRAY)
    images = [gray]
    for t in thresholds:
        mask = tissue_mask(thumbnail, t)
        # Flip for a nicer image.
        mask = 1 - mask
        mask = mask*255
        images.append(mask.astype(np.uint8))
    images = [images[i:i + 4] for i in range(0, len(images), 4)]
    rows = []
    for row in images:
        while len(row) != 4:
            row.append(np.ones(row[0].shape)*255)
        rows.append(np.hstack(row))
    summary = Image.fromarray(np.vstack(rows).astype('uint8'))
    l = ['original'] + thresholds
    print('Thresholds:\n')
    for row in [l[i:i + 4] for i in range(0, len(l), 4)]:
        [print(str(x).center(8), end='') for x in row]
        print('')
    return resize(summary, max_pixels)


#################################
### Functions for Dearrayer() ###
#################################

def detect_spots(
        mask: np.ndarray,
        min_area_multiplier: float,
        max_area_multiplier: float,
        kernel_size: Tuple[int, int],):
    """ Detect TMA spots from a thumbnail image.

    How: Detect tissue mask -> clean up non-TMA stuff -> return mask.

    Arguments:
        min_area_multiplier: 
            median_spot_area * min_area_multiplier
        max_area_multiplier: 
            median_spot_area * max_area_multiplier
        kernel_size: 
            Sometimes the default doesn't work for large/small thumbnails.
    """
    # Opening to remove small shit.
    opening = cv2.morphologyEx(mask, cv2.MORPH_OPEN, kernel_size, iterations=3)
    # Dilate.
    mask = cv2.dilate(mask, np.ones(kernel_size, np.uint8), iterations=3)
    # Remove too small/large spots.
    contours, __ = cv2.findContours(
        mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    areas = np.array([cv2.contourArea(x) for x in contours])
    # Define min and max values.
    min_area = np.median(areas)*min_area_multiplier
    if max_area_multiplier is not None:
        max_area = np.mean(areas)*max_area_multiplier
    else:
        max_area = np.max(areas)
    idx = (areas >= min_area) & (areas <= max_area)
    contours = [contours[i] for i in range(len(idx)) if idx[i]]
    # Draw new mask.
    new_mask = np.zeros(mask.shape, dtype="uint8")
    for i, cnt in enumerate(contours):
        cv2.drawContours(new_mask, [contours[i]], -1, 1, -1)
    return new_mask


def get_spots(spot_mask: np.ndarray, downsample: int):
    """Orders the spots on a spot_mask taking into consideration empty spots."""
    # Collect bounding boxes.
    contours, _ = cv2.findContours(
        spot_mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
    if len(contours) == 0:
        return None, None
    # The format is (x,y,w,h)
    boxes = [cv2.boundingRect(cnt) for cnt in contours]
    boxes = np.array(boxes) * downsample
    # Get centroid of each bounding box.
    coords = np.zeros(boxes[:, :2].shape)
    for i, cnt in enumerate(contours):
        M = cv2.moments(cnt)
        centroid_x = int(M["m10"] / M["m00"])
        centroid_y = int(M["m01"] / M["m00"])
        coords[i, :] = [centroid_x, centroid_y]
    # Rotate coordinates.
    theta = get_theta(coords)
    coords = rotate_coordinates(coords, theta)
    # Detect optimal number of rows and columns.
    cols = get_optimal_cluster_size(coords[:, 0])
    rows = get_optimal_cluster_size(coords[:, 1])
    # Cluster.
    col_labels = hierachial_clustering(coords[:, 0], n_clusters=cols)
    row_labels = hierachial_clustering(coords[:, 1], n_clusters=rows)
    # Detect cluster means.
    x_means = [coords[col_labels == i, 0].mean() for i in range(cols)]
    y_means = [coords[row_labels == i, 1].mean() for i in range(rows)]
    # Change label numbers to correct order (starting from top-left).
    for i in range(cols):
        new_label = np.arange(cols)[np.argsort(x_means) == i]
        col_labels[col_labels == i] = -new_label
    col_labels *= -1
    for i in range(rows):
        new_label = np.arange(rows)[np.argsort(y_means) == i]
        row_labels[row_labels == i] = -new_label
    row_labels *= -1
    labels = list(zip(col_labels, row_labels))
    # Collect numbers.
    numbers = np.zeros(len(coords)).astype(np.str)
    i = 1
    for r in range(rows):
        for c in range(cols):
            idx = [x == (c, r) for x in labels]
            if sum(idx) == 1:
                numbers[idx] = i
            elif sum(idx) > 1:
                numbers[idx] = [f'{i}_{ii}' for ii in range(sum(idx))]
            i += 1
    return numbers, boxes


def get_theta(coords: np.ndarray) -> float:
    """Detect rotation from centroid coordinates and return angle in radians."""
    n = len(coords)
    thetas = []
    for r in range(n):
        for c in range(n):
            x1, y1 = coords[r, :]
            x2, y2 = coords[c, :]
            thetas.append(np.rad2deg(np.arctan2(y2 - y1, x2 - x1)))
    # We want deviations from 0 so divide corrections.
    corr = np.array([0, 45, 90, 135, 180])
    for i, theta in enumerate(thetas):
        sign = np.sign(theta)
        idx = np.abs(np.abs(theta)-corr).argmin()
        thetas[i] = theta-sign*corr[idx]
    # Finally return most common angle
    values, counts = np.unique(np.round(thetas), return_counts=True)
    theta = values[counts.argmax()]
    return np.radians(theta)


def rotate_coordinates(coords: np.ndarray, theta: float) -> np.ndarray:
    """Rotate coordinates with given theta."""
    c, s = np.cos(theta), np.sin(theta)
    R = np.array(((c, -s), (s, c)))
    return coords @ R


def get_optimal_cluster_size(X: np.ndarray) -> int:
    """Find optimal cluster size for dataset X."""
    if len(X.shape) < 2:
        X = X.reshape(-1, 1)
    sil = []
    for n in range(2, X.shape[0]):
        clust = AgglomerativeClustering(n_clusters=n, linkage='ward')
        clust.fit(X)
        sil.append(silhouette_score(X, clust.labels_))
    return np.argmax(sil)+2


def hierachial_clustering(
        X: np.ndarray,
        n_clusters: int,
        linkage: str = 'ward'
) -> np.ndarray:
    if len(X.shape) < 2:
        X = X.reshape(-1, 1)
    clust = AgglomerativeClustering(n_clusters=n_clusters, linkage=linkage)
    clust.fit(X)
    return clust.labels_



import os
import itertools
import multiprocessing as mp
from functools import partial
from typing import List
import logging

from tqdm import tqdm
import cv2
import numpy as np
from PIL import Image, ImageDraw
from aicspylibczi import CziFile
from shapely.geometry import Polygon, MultiPolygon, mapping

__all__ = [
    'OpenSlideCzi'
]

# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class OpenSlideCzi(object):
    """
    Open a czi-file like one of your french OpenSlide readers.

    Args:
        slide_path (str): Path to the slide.

    Raises:
        NotImplementedError: czi-file is non-mosaic.
    """

    def __init__(self, slide_path: str):
        self.slide_path = slide_path
        reader = CziFile(slide_path)
        if not reader.is_mosaic():
            raise NotImplementedError(
                'This class has only been defined for mosaic czi files. '
                'You should be able to simply convert the non-mosaic czi-file '
                'into a format that can be opened with openslide.'
            )
        self.bboxes = np.array(reader.mosaic_scene_bounding_boxes())
        self.data_mask = self._get_data_mask(self.bboxes)
        self.shape = reader.read_mosaic_size()
        self.dimensions = self.shape[2:]
        self.region_mask = self._get_region_mask(self.shape)

    def _get_data_mask(self, bboxes):
        polys = []
        for x, y, w, h in bboxes:
            polys.append(
                Polygon([[x, y], [x+w, y], [x+w, y+h], [x, y+h], [x, y]]))
        return MultiPolygon(polys).buffer(0)

    def _get_region_mask(self, shape):
        x, y, w, h = shape
        mask = Polygon([[x, y], [x+w, y], [x+w, y+h], [x, y+h], [x, y]])
        return mask

    def _get_all_coordinates(self, bboxes, width, overlap=0.0):
        """Return all coordinates."""
        stop_x = bboxes[:, 0].max()
        stop_y = bboxes[:, 1].max()
        x = [bboxes[:, 0].min()]
        y = [bboxes[:, 1].min()]
        overlap_px = int(width * overlap)
        while x[-1] < stop_x:
            x.append(x[-1] + width - overlap_px)
        x = x[:-1]
        while y[-1] < stop_y:
            y.append(y[-1] + width - overlap_px)
        y = y[:-1]
        blocks = (len(x), len(y))
        coordinates = list(itertools.product(x, y))
        return coordinates, blocks

    def get_tiles_with_data(self, width, overlap):
        coords, __ = self._get_all_coordinates(self.bboxes, width, overlap)
        # Wrap function.
        func = partial(check_tile, **{
            'data_mask': self.data_mask,
            'region_mask': self.region_mask,
            'width': width,
        })
        # Load tiles.
        filtered = []
        with mp.Pool(processes=os.cpu_count() - 1) as p:
            for result in tqdm(
                p.imap(func, coords),
                total=len(coords),
                desc='Filtering tiles',
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                filtered.append(result)
        filtered = list(filter(lambda x: x is not None, filtered))
        return filtered

    def read_region(self, location, scale, size):
        """Return a PIL.Image containing the contents of the region.

        Args:
            location: (x, y) tuple giving the top left pixel in the level 0
                      reference frame.
            level:    the level number.
            size:     (width, height) tuple giving the region size.
        """
        # Fix coordinates to czi's shitty negative ones...
        x, y = location
        x = self.shape[0] + x
        y = self.shape[1] + y
        location = (0, (x, y))
        width = size[0] * (2 ** scale)
        __, tile = load_tile(
            coords=location,
            slide_path=self.slide_path,
            data_mask=self.data_mask,
            region_mask=self.region_mask,
            width=width,
            downsample=2**scale,
            fast=False
        )
        return Image.fromarray(tile.astype(np.uint8))

    def get_thumbnail(self, downsample: int, width: int = 4096,
                      fast: bool = False) -> Image.Image:
        """
        Generate a thumbnail image with given downsample.

        Args:
            downsample (int): Desired downsample for the thumbnail.
            width (int, optional): Width of each tile during construction.
                Defaults to 4096.
            fast (bool, optional): Use lower pyramid levels in the czi-file to 
                generate each tile. If the slide contains two or more scenes 
                then some parts of the thumbnail can be missing. It is fast as
                fuck though so idk... Defaults to False.

        Returns:
            [Image.Image]: Thumbnail image.
        """
        # Load ALL coordinates.
        coordinates, blocks = self._get_all_coordinates(self.bboxes, width)
        coordinates = list(enumerate(coordinates))
        # Wrap function.
        func = partial(load_tile, **{
            'slide_path': self.slide_path,
            'data_mask': self.data_mask,
            'region_mask': self.region_mask,
            'width': width,
            'downsample': downsample,
            'fast': fast,
        })
        # Load tiles.
        tiles = []
        with mp.Pool(processes=os.cpu_count() - 1) as p:
            for result in tqdm(
                p.imap(func, coordinates),
                total=len(coordinates),
                desc='Generating thumbnail',
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                tiles.append(result)
        tiles.sort()
        # Check tiles...
        if len(tiles) == 0:
            logger.warning('No tiles found!')
            return None
        # Collect each column seperately and mash them together.
        all_columns = []
        col = []
        for i, tile in tiles:
            if i % blocks[1] == 0 and i != 0:
                all_columns.append(np.vstack(col))
                col = []
            col.append(np.array(tile))
        thumbnail = np.hstack(all_columns)
        # Turn into into pillow image.
        thumbnail = Image.fromarray(thumbnail.astype(np.uint8)).convert('RGB')
        return thumbnail


def polygon_to_mask(poly, x, y, width, downsample):
    """Turn polygon to tile mask."""
    mask = Image.new('L', (int(width/downsample), int(width/downsample)), 0)
    mapp = mapping(poly)
    if mapp['type'] == 'Polygon' or mapp['type'] == 'MultiPolygon':
        coords = mapp['coordinates']
    elif mapp['type'] == 'GeometryCollection':
        coords = []
        for key, val in mapp.items():
            if key == 'geometries':
                for d in val:
                    if d['type'] == 'Polygon':
                        coords.append(d['coordinates'])
    else:
        return mask
    if len(coords) > 1:
        coords = [item for sublist in coords for item in sublist[0]]
    else:
        coords = coords[0]
    if len(coords) > 2 and isinstance(coords[0], tuple):
        coords = [((c[0] - x)/downsample, (c[1] - y)/downsample)
                  for c in coords]
        ImageDraw.Draw(mask).polygon(coords, outline=1, fill=1)
    mask = np.array(mask)
    return mask


def check_tile(xy, data_mask, region_mask, width):
    __, (x, y) = xy 
    tile = Polygon([
        [x, y],
        [x+width, y],
        [x+width, y+width],
        [x, y+width],
        [x, y]
    ])
    d_perc = tile.intersection(data_mask).area/tile.area
    r_perc = tile.intersection(region_mask).area/tile.area
    if d_perc == 1 and r_perc == 1:
        return xy
    else:
        return None


def load_tile(coords, slide_path, data_mask,
              region_mask, width, downsample, fast):
    """Load a tile from the czi-file (parallizeable)."""
    # Load reader.
    reader = CziFile(slide_path)
    # Unpack coords.
    i, (x, y) = coords
    # Define final out_shape.
    out_shape = (int(width/downsample), int(width/downsample))
    # Get data and region percentages.
    tile_mask = Polygon(
        [[x, y], [x+width, y], [x+width, y+width], [x, y+width], [x, y]])
    intersection = tile_mask.intersection(data_mask)
    data_perc = intersection.area/tile_mask.area
    region_perc = tile_mask.intersection(region_mask).area/tile_mask.area
    if data_perc < 0.05 or region_perc < 1:
        # Return empty image.
        tile = np.ones(out_shape + (3,)) * 255
    else:
        # Load tile.
        bbox = (x, y, width, width)
        if fast:
            tile = reader.read_mosaic(bbox, C=0, scale_factor=1/downsample)
        else:
            tile = reader.read_mosaic(bbox, C=0)
        if tile.shape[0] == 1:
            # Something is wrong with the tile...
            tile = np.ones(out_shape + (3,)) * 255
        else:
            tile = np.moveaxis(tile, 0, 2)
            tile = cv2.resize(tile, out_shape, cv2.INTER_LANCZOS4)
            if data_perc < 1:
                mask = polygon_to_mask(intersection, x, y, width, downsample)
                mask = cv2.resize(mask, out_shape, cv2.INTER_LANCZOS4)
                tile[mask == 0] = 255
    tile = tile.astype(np.uint8)
    return i, tile



from typing import Union, List, Dict, Any, Tuple

import numpy as np
import cv2
from PIL import Image

__all__ = [
    'tissue_mask',
    'HSV_quantiles',
    'RGB_quantiles',
    'data_loss',
    'laplacian_variance',
    'sharpness',
    'preprocess',
    'sliding_window',
    'PIL_to_array',
    'array_to_PIL',
    'mask_to_PIL',
]


def PIL_to_array(image: Image.Image) -> np.ndarray:
    """
    Convert Pillow image to numpy array.

    Args:
        image (Image.Image): Input image.

    Raises:
        TypeError: Invalid input type for ``image``.

    Returns:
        np.ndarray: ``image`` as an numpy array.
    """
    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        return np.array(image)
    else:
        raise TypeError('Excpected {} not {}.'.format(
            Image.Image, type(image)))


def array_to_PIL(image: np.ndarray) -> Image.Image:
    """
    Convert numpy array to Pillow image,

    [extended_summary]

    Args:
        image (np.ndarray): Input image.

    Raises:
        TypeError: Invalid input type for ``image``.

    Returns:
        Image.Image: ``image`` as a Pillow image.
    """
    if isinstance(image, np.ndarray):
        return Image.fromarray(image.astype(np.uint8))
    else:
        raise TypeError('Excpected {} not {}.'.format(np.ndarray, type(image)))

    """"""


def mask_to_PIL(mask: np.ndarray) -> Image.Image:
    """
    Normalize a numpy mask between 0 and 255 and convert to PIL image.

    Args:
        mask (np.ndarray): Mask as an numpy array.

    Raises:
        TypeError: Invalid mask type.

    Returns:
        Image.Image: Mask image.
    """
    if isinstance(mask, np.ndarray):
        # Normalize between 0-255.
        if mask.max() != 0:
            mask = (mask/mask.max()) * 255
        return Image.fromarray(mask.astype(np.uint8))
    else:
        raise TypeError('Excpected {} not {}.'.format(np.ndarray, type(mask)))


def tissue_mask(
        image: Union[np.ndarray, Image.Image],
        threshold: int = None,
        blur_kernel: Tuple[int, int] = (5, 5),
        blur_iterations: int = 1,
        return_threshold: bool = False) -> np.ndarray:
    """
    Generate a tissue mask for image.

    Two methods are implemented.

    Otsu's binarization:
        Otsu's method is used to find an optimal threshold by minimizing the 
        weighted within-class variance. Due to this, a relatively high 
        threshold for tissue detection is often selected and actual tissue
        is misclassified as background. Binarization is also forced even for 
        tiles with only background, causing the detection of non-existent 
        tissue.

    Adaptive gaussian thresholding:
        Requires a threshold to be given but performs better than Otsu's method.
        This is automatically implemented if a threshold is given.

    Args:
        image (Union[np.ndarray, Image.Image]): Input image.
        threshold (int, optional): Threshold for tissue detection (see the 
            method explanation above). Defaults to None.
        blur_kernel (Tuple[int, int], optional): Kernel to be used in Gaussian 
            Blur. Set to None to disable. Defaults to (5, 5).
        blur_iterations (int, optional): How many iterations to blur with 
            kernel. Defaults to 1.
        return_threshold (bool, optional): Whether to return the used threshold
            in the case of Otsu's method. Defaults to False.

    Raises:
        TypeError: Invalid type for ``image`` or ``threshold``.

    Returns:
        np.ndarray: tissue mask.
    """

    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    # Turn RGB to GRAY
    gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    # Blur if asked.
    if blur_kernel is not None:
        gray = cv2.GaussianBlur(gray, blur_kernel, blur_iterations)
    # Then do thresholding.
    if threshold is None:
        thresh, mask = cv2.threshold(
            src=gray,
            thresh=None,
            maxval=1,
            type=cv2.THRESH_BINARY+cv2.THRESH_OTSU
        )
        mask = 1 - mask
    else:
        try:
            threshold = int(threshold)
        except:
            raise TypeError(f'Excpected {int} not {type(threshold)}.')
        thresh, mask = cv2.threshold(
            src=gray,
            thresh=threshold,
            maxval=1,
            type=cv2.ADAPTIVE_THRESH_GAUSSIAN_C
        )
    if return_threshold:
        return int(thresh), mask
    else:
        return mask


def RGB_quantiles(
        image: Union[np.ndarray, Image.Image],
        quantiles: List[float] = [.01, .05, 0.1,
                                  0.25, 0.5, 0.75, 0.9, 0.95, 0.99],
        gray: np.ndarray = None,
        mask: np.ndarray = None,
        resize: int = None,
        threshold: int = None) -> Dict[int, int]:
    """
    Measure color channel quantiles.

    Useful in the detection of misclassified tissue and artifacts.

    Args:
        image (Union[np.ndarray, Image.Image]): Input image.
        quantiles (List[float], optional): Quantiles to be collected. Defaults 
            to [.01, .05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99].
        gray (np.ndarray, optional): Grayscale image of the input image. Will be
            generated if not defined. Defaults to None.
        mask (np.ndarray, optional): Tissue mask for the input image. 
            Will be generated if not defined. Defaults to None.
        resize (int, optional): Resize the image to resize x resize. The 
            function can become really slow with large images as we have to sort
            every pixel in the image. In these situations just use this
            option. Defaults to None.
        threshold (int, optional): For tissue_mask() function. Ignored if mask 
            is defined. Defaults to None.

    Raises:
        TypeError: Invalid type for ``image``.

    Returns:
        Dict[int, int]: Dictionary of color channel quantiles.
    """

    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    if resize is not None:
        image = cv2.resize(image, (resize, resize), cv2.INTER_LANCZOS4)
    if mask is None:
        mask = tissue_mask(image, threshold=threshold)
    elif mask.shape != image.shape[:2]:
        mask = cv2.resize(mask, (resize, resize), cv2.INTER_LANCZOS4)
    if mask.sum() == 0:
        # No tissue, return empty dict
        return {}
    # Collect channels and sort.
    red = np.sort(image[mask == 1, 0])
    green = np.sort(image[mask == 1, 1])
    blue = np.sort(image[mask == 1, 2])
    # Collect quantiles.
    red = [np.quantile(red, q) for q in quantiles]
    green = [np.quantile(green, q) for q in quantiles]
    blue = [np.quantile(blue, q) for q in quantiles]
    keys = (
        [f'red_{x}' for x in quantiles] +
        [f'green_{x}' for x in quantiles] +
        [f'blue_{x}' for x in quantiles]
    )
    results = dict(zip(keys, red + green + blue))
    return results

    """HSV channel quantiles.

    Useful in the detection of misclassified tissue and artifacts.

    Arguments:
        image: 
            Input image.
        mask: 
            Tissue mask for the input image. Will be generated if not defined.
        quantiles: 
            The quantiles of hue, sat and value values for tissue areas.
        resize: 
            Resize the image to resize x resize. The function can become 
            really slow with large images, in these situations just use this
            option.
        threshold: 
            For tissue_mask() function. Ignored if mask is defined.

    Return:
        dict: A dictionary of the quantiles of hue, sat and value values for 
            tissue areas.
    """


def HSV_quantiles(
        image: Union[np.ndarray, Image.Image],
        quantiles: List[float] = [.01, .05, 0.1,
                                  0.25, 0.5, 0.75, 0.9, 0.95, 0.99],
        mask: np.ndarray = None,
        resize: int = None,
        threshold: int = None) -> Dict[int, int]:
    """
    Measure HSV channels quantiles.

    Useful in the detection of misclassified tissue and artifacts.

    Args:
        image (Union[np.ndarray, Image.Image]): Input image.
        quantiles (List[float], optional): Quantiles to be collected. Defaults 
            to [.01, .05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99].
        mask (np.ndarray, optional): Tissue mask for the input image. 
            Will be generated if not defined. Defaults to None.
        resize (int, optional): Resize the image to resize x resize. The 
            function can become really slow with large images as we have to sort
            every pixel in the image. In these situations just use this
            option. Defaults to None.
        threshold (int, optional): For tissue_mask() function. Ignored if mask 
            is defined. Defaults to None.

    Raises:
        TypeError: Invalid type for ``image``.

    Returns:
        Dict[int, int]: Dictionary of HSV channels quantiles.
    """
    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    if resize is not None:
        image = cv2.resize(image, (resize, resize), cv2.INTER_LANCZOS4)
    if mask is None:
        mask = tissue_mask(image, threshold=threshold)
    elif mask.shape != image.shape[:2]:
        mask = cv2.resize(mask, (resize, resize), cv2.INTER_LANCZOS4)
    if mask.sum() == 0:
        # No tissue, return empty dict
        return {}
    # Collect channels and sort.
    HSV = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)
    hue = np.sort(HSV[mask == 1, 0])
    sat = np.sort(HSV[mask == 1, 1])
    val = np.sort(HSV[mask == 1, 2])
    # Collect quantiles.
    hue = [np.quantile(hue, q) for q in quantiles]
    sat = [np.quantile(sat, q) for q in quantiles]
    val = [np.quantile(val, q) for q in quantiles]
    keys = (
        [f'hue_{x}' for x in quantiles] +
        [f'sat_{x}' for x in quantiles] +
        [f'val_{x}' for x in quantiles]
    )
    results = dict(zip(keys, hue + sat + val))
    return results


def data_loss(image: Union[np.ndarray, Image.Image]) -> Dict[float, float]:
    """
    Detect data loss.

    Calculates the percentage of completely white and black pixels.

    Args:
        image (Union[np.ndarray, Image.Image]): Input image.

    Raises:
        TypeError: Invalid type for ``image``.

    Returns:
        Dict[float, float]: Percentages of completely black (0) and white (255)
            pixels.
    """
    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    if len(image.shape) > 2:
        gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    else:
        gray = image
    return {
        'black_pixels': gray[gray == 0].size/gray.size,
        'white_pixels': gray[gray == 255].size/gray.size
    }

def laplacian_variance(image: Union[np.ndarray, Image.Image]):
    """
    Return the laplacian variance of the image.

    Args:
        image (Union[np.ndarray, Image.Image]): input image.

    Raises:
        TypeError: Invalid type for ``image``.
    """
    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    # Laplacian variance is defined for greyscale images.
    if len(image.shape) > 2:
        gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    else:
        gray = image
    return cv2.Laplacian(gray, cv2.CV_32F).var()

def sharpness(
        image: Union[np.ndarray, Image.Image],
        divider: int = 2,
        reduction: Union[str, List[str]] = 'max') -> dict:
    """
    Sharpness detection with Laplacian variance.

    Divides the image into 9, 25, 49, ... tiles with 50% overlap based on 
    ``divider``, calculates the Laplacian sharpness for each tile and returns
    the value based on ``reduction``.

    Args:
        image (Union[np.ndarray, Image.Image]): Input image.
        divider (int, optional): Divider argument for the ``sliding_window()`` 
            function. Defaults to 2.
        reduction (Union[str, List[str]], optional): Reduction method(s) for the
            Laplacian variance values for each window. Defaults to 'max'.

    Raises:
        TypeError: Invalid type for ``image``.

    Returns:
        dict: Laplacian variance values. 
    """
    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    # Laplacian variance is defined for greyscale images.
    if len(image.shape) > 2:
        gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    else:
        gray = image
    values = []
    for window in sliding_window(image):
        values.append(cv2.Laplacian(window, cv2.CV_32F).var())
    # Then reduction(s).
    if isinstance(reduction, str):
        reduction = [reduction]
    results = dict(zip(
        ['sharpness_'+method for method in reduction],
        [reduce_values(values, method) for method in reduction]
    ))
    return results


def reduce_values(values: list, method: str):
    """Reduce values of values with given method."""
    allowed_methods = ['max', 'median', 'mean', 'min']
    if method not in allowed_methods:
        raise ValueError('Reduction {} not recognised. Select from {}.'.format(
            reduction, allowed_methods
        ))
    if method == 'max':
        return np.max(values)
    elif method == 'median':
        return np.median(values)
    elif method == 'mean':
        return np.mean(values)
    elif method == 'min':
        return np.min(values)


    """
    Sliding window with 0.5 overlap.

    :param image: Input image.
    :type image: Union[np.ndarray, Image.Image]
    :param divider: Window size is defined as min(height,width)/divider, where
        divider defaults to 2. For square images, divider values will produce:
            1: original image
            2: 3x3=9 windows
            3: 5x5=25 windows
            4: 7x7=49 windows
    :type divider: int, optional
    :raises TypeError: If image is in a wrong format.
    :return: List of window images.
    :rtype: List[np.ndarray]
    """

def sliding_window(
        image: Union[np.ndarray, Image.Image],
        divider: int = 2) -> List[np.ndarray]:
    """
    Sliding window with 0.5 overlap.


    Args:
        image (Union[np.ndarray, Image.Image]): Input image.
        divider (int, optional): Window size is defined as 
            min(height,width)/divider, where divider defaults to 2. For square 
            images, divider values will produce:

            1: original image
            2: 3x3=9 windows
            3: 5x5=25 windows
            4: 7x7=49 windows.

    Raises:
        TypeError: Invalid type for ``image``.

    Returns:
        List[np.ndarray]: List of window images.
    """
    if isinstance(image, Image.Image):
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    if divider < 2:
        return [image]
    w = int(min(x/divider for x in image.shape[:2]))
    rows, cols = [int(x/(w/2)-1) for x in image.shape[:2]]
    windows = []
    for row in range(rows):
        for col in range(cols):
            r = int(row*w/divider)
            c = int(col*w/divider)
            windows.append(image[r:r+w, c:c+w])
    return windows


def preprocess(
        image: Union[np.ndarray, Image.Image],
        threshold: int = None,
        resize: int = 64,
        quantiles: List[float] = [.01, .05, 0.1,
                                  0.25, 0.5, 0.75, 0.9, 0.95, 0.99],
        reduction: Union[str, List[str]] = ['max', 'median', 'mean', 'min']
) -> dict:
    """
    Basic preprocessing metrics for a histological image.

    Args:
        image (Union[np.ndarray, Image.Image]):  Input image.
        threshold (int, optional): Threshold for tissue detection. If not 
            defined Otsu's binarization will be used (which) may fail for images
            with data loss or only background. Defaults to None.
        resize (int, optional): For artifact() function. Defaults to 64.
        quantiles (List[float], optional): For HSV_quantiles() and RGB_quantiles
            functions. Defaults to 
            [.01, .05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99].
        reduction (Union[str, List[str]], optional): Reduction methods for
            sharpness() function. Defaults to ['max', 'median', 'mean', 'min'].

    Raises:
        TypeError: Invalid type for ``image``.

    Returns:
        dict: Dictionary of basic preprocessing metrics.
    """

    if isinstance(image, Image.Image):
        if image.mode != 'RGB':
            image = image.convert('RGB')
        image = np.array(image, dtype=np.uint8)
    elif isinstance(image, np.ndarray):
        image = image.astype(np.uint8)
    else:
        raise TypeError('Excpected {} or {} not {}.'.format(
            np.ndarray, Image.Image, type(image)
        ))
    # Initialize results and other shit.
    results = {}
    gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
    mask = tissue_mask(image, threshold=threshold)
    # Background percentage.
    results['background'] = (mask == 0).sum()/mask.size
    # Sharpness.
    results.update(sharpness(gray, reduction=reduction))
    # Data loss.
    results.update(data_loss(gray))
    # Artifacts.
    small_img = cv2.resize(image, (resize, resize), cv2.INTER_LANCZOS4)
    small_mask = cv2.resize(mask, (resize, resize), cv2.INTER_LANCZOS4)
    results.update(HSV_quantiles(
        small_img, mask=small_mask, quantiles=quantiles))
    results.update(RGB_quantiles(
        small_img, mask=small_mask, quantiles=quantiles))

    return results



import setuptools

exec(open('histoprep/_version.py').read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="histoprep",
    version=__version__,
    author="jopo666",
    author_email="jopo@birdlover.com",
    description="Preprocessing module for large histological images.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jopo666/HistoPrep",
    packages=setuptools.find_packages(include=['histoprep','histoprep.*']),
    install_requires=[
        'opencv-python==4.5.1.48',
        'openslide-python==1.1.2',
        'pandas==1.2.1',
        'Pillow==8.0.0',
        'seaborn==0.11.0',
        'numpy==1.19.2',
        'tqdm==4.60.0',
        'aicspylibczi==2.8.0',
        'shapely==1.7.1',
        'scikit-learn==0.24.1',
        'ipywidgets==7.6.3',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords='image-analysis preprocessing histology openslide',
    python_requires='>=3.8',
)


import os
from os.path import dirname, join, basename, exists
from typing import List, Tuple, Callable
import itertools
import multiprocessing as mp
from functools import partial
import warnings
import logging

import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm
from PIL import Image, ImageDraw, ImageFont

from openslide import OpenSlide

from ._functional import (
    get_thumbnail,
    get_downsamples,
    try_thresholds,
    resize,
    detect_spots,
    get_spots,
    get_theta
)
from .preprocess.functional import preprocess, tissue_mask
from .helpers._utils import remove_extension, remove_images, flatten

__all__ = [
    'Dearrayer'
]

# Define logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Dearrayer(object):
    """
    Cut TMA spots from a TMA array slide.

    Args:
        slide_path (str): Path to the TMA slide array. All formats that are
            supported by openslide can be used.
        threshold (int, optional): Threshold value for tissue detection.
            Can be left undefined, in which case Otsu's binarization is used.
            This is not recommended! Values can easily be searched with
            Cutter.try_thresholds() function. Defaults to None.
        downsample (int, optional): Downsample used for the thumbnail. The user
            might have to tweak this value depending on the magnification of the
            slide. The TMA spot detection method is optimized for downsamlpe=64
            when the magnification is 20x. If no spots are found, try adjusting
            the downsample or tweak the spot detection variables with
            ``Dearrayer.try_spot_mask()`` function.  Defaults to 64.
        min_area_multiplier (float, optional): Remove all detected contours that
            have an area smaller than ``median_area*min_area_multiplier``.
            Defaults to 0.2.
        max_area_multiplier (float, optional): Remove all detected contours that
            have an area larger than ``median_area*max_area_multiplier``.
            Defaults to None.
        kernel_size (Tuple[int], optional): Kernel size used during spot
            detection. Defaults to (8, 8).
        create_thumbnail (bool, optional):  Create a thumbnail if downsample is
            not available. Defaults to False.

    Raises:
        IOError: slide_path not found.
        ValueError: downsample is not available and create_thumbnail=False.
    """

    def __init__(
            self,
            slide_path: str,
            threshold: int = None,
            downsample: int = 64,
            min_area_multiplier: float = 0.2,
            max_area_multiplier: float = None,
            kernel_size: Tuple[int] = (8, 8),
            create_thumbnail: bool = False):
        super().__init__()
        # Define openslide reader.
        if not exists(slide_path):
            raise IOError(f'{slide_path} not found.')
        self.openslide_reader = OpenSlide(slide_path)
        # Assing basic stuff that user can see/check.
        self.slide_path = slide_path
        self.slide_name = remove_extension(basename(slide_path))
        self.dimensions = self.openslide_reader.dimensions
        self.downsample = downsample
        self.threshold = threshold
        self.min_area_multiplier = min_area_multiplier
        self.max_area_multiplier = max_area_multiplier
        self.kernel_size = kernel_size
        self._spots_saved = False
        # Get spots.
        self._thumbnail = get_thumbnail(
            slide_path=self.slide_path,
            downsample=self.downsample,
            create_thumbnail=create_thumbnail
        )
        if self._thumbnail is None:
            # Downsample not available.
            raise ValueError(
                f'Thumbnail not available for downsample {self.downsample}. '
                'Please set create_thumbnail=True or select downsample from\n\n'
                f'{self._downsamples()}'
            )
        self.threshold, self._tissue_mask = tissue_mask(
            image=self._thumbnail,
            threshold=self.threshold,
            return_threshold=True
        )
        self._spot_mask = detect_spots(
            mask=self._tissue_mask,
            min_area_multiplier=self.min_area_multiplier,
            max_area_multiplier=self.max_area_multiplier,
            kernel_size=self.kernel_size,
        )
        self._numbers, self._bounding_boxes = get_spots(
            spot_mask=self._spot_mask,
            downsample=self.downsample,
        )
        if self._numbers is None or self._bounding_boxes is None:
            logger.warning(
                'No spots detected from the slide! Please try and adjust, '
                'the kernel_size, min_area_multiplier and max_area_multiplier '
                'parameters using the dearrayer.try_spot_mask() function.'
            )
            self.spot_metadata = None
            self._annotate()
        else:
            self.spot_metadata = pd.DataFrame(
                np.hstack((self._numbers.reshape(-1, 1), self._bounding_boxes)))
            self.spot_metadata.columns = [
                'number', 'x', 'y', 'width', 'height']
            self._annotate()
            if len([x for x in self._numbers if '_' in x]) > 0:
                logger.warning(
                    'Some spots were assinged the same number. Please check the '
                    f'annotated thumbnail for slide {self.slide_name}.'
                )

    def __repr__(self):
        return self.__class__.__name__ + '()'

    def __len__(self):
        return len(self._bounding_boxes)

    def summary(self):
        """Returns a summary of the dearraying process."""
        print(self._summary())

    def _summary(self, cut=False):
        summary = (
            f"{self.slide_name}"
            f"\n  Number of TMA spots: {len(self._bounding_boxes)}"
            f"\n  Downsample: {self.downsample}"
            f"\n  Threshold: {self.threshold}"
            f"\n  Min area multiplier: {self.min_area_multiplier}"
            f"\n  Max area multiplier: {self.max_area_multiplier}"
            f"\n  Kernel size: {self.kernel_size}"
            f"\n  Dimensions: {self.dimensions}"
        )
        if cut:
            summary += (
                f"\n  Tile width: {self.width}"
                f"\n  Tile overlap: {self.overlap}"
                f"\n  Max background: {self.max_background}"
            )
        return summary

    def get_thumbnail(self, max_pixels: int = 1_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the thumbnail for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: Thumbnail.
        """
        return resize(self._thumbnail, max_pixels)

    def get_annotated_thumbnail(self,
                                max_pixels: int = 5_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the annotated thumbnail for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image
                has less than max_pixles pixels. Defaults to 5_000_000.

        Returns:
            Image.Image: Annotated thumbnail.
        """
        return resize(self._annotated_thumbnail, max_pixels)

    def get_tissue_mask(self, max_pixels: int = 1_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the tissue mask for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: Tissue mask.
        """
        mask = self._tissue_mask
        # Flip for a nicer image.
        mask = 1 - mask
        mask = mask/mask.max()*255
        mask = Image.fromarray(mask.astype(np.uint8))
        return resize(mask, max_pixels)

    def get_spot_mask(self, max_pixels: int = 1_000_000) -> Image.Image:
        """
        Returns an Pillow Image of the TMA spot mask for inspection.

        Args:
            max_pixels (int, optional): Downsample the image until the image
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: Spot mask.
        """
        mask = self._spot_mask
        # Flip for a nicer image.
        mask = 1 - mask
        mask = mask/mask.max()*255
        mask = Image.fromarray(mask.astype(np.uint8))
        return resize(mask, max_pixels)

    def _annotate(self):
        """Draw bounding boxes and numbers to the thumbnail."""
        fontsize = (self.spot_metadata.width.median()/6000)*70/self.downsample
        self._annotated_thumbnail = self._thumbnail.copy()
        if self.spot_metadata is None:
            return
        else:
            annotated = ImageDraw.Draw(self._annotated_thumbnail)
            # Bounding boxes.
            for i in range(len(self)):
                x, y, w, h = self._bounding_boxes[i]/self.downsample
                annotated.rectangle(
                    [x, y, x+w, y+h], outline='red', width=round(fontsize*5))
            arr = np.array(self._annotated_thumbnail)
            # Numbers.
            for i in range(len(self)):
                x, y, w, h = self._bounding_boxes[i]/self.downsample
                arr = cv2.putText(
                    arr,
                    str(self._numbers[i]),
                    (int(x+10), int(y-10+h)),
                    cv2.FONT_HERSHEY_SCRIPT_SIMPLEX,
                    fontsize,
                    (0, 0, 255),
                    round(fontsize*3),
                    cv2.LINE_AA
                )
            self._annotated_thumbnail = Image.fromarray(arr)

    def try_thresholds(
            self,
            thresholds: List[int] = [250, 240, 230,
                                     220, 200, 190, 180, 170, 160, 150, 140],
            max_pixels=1_000_000) -> Image.Image:
        """
        Try out different thresholds for tissue detection.

        The function prepares tissue masks with given thresholds and slaps them
        all together in one summary image.

        Args:
            thresholds (List[int], optional): Thresholds to try. Defaults to
                [250, 240, 230, 220, 200, 190, 180, 170, 160, 150, 140].
            max_pixels (int, optional): Downsample the image until the image
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: [description]
        """
        return try_thresholds(thumbnail=self._thumbnail, thresholds=thresholds)

    def try_spot_mask(
            self,
            min_area_multiplier: float = 0.1,
            max_area_multiplier: float = 2,
            kernel_size: Tuple[int] = (5, 5),
            max_pixels: int = 1_000_000) -> Image.Image:
        """
        Try out different values for TMA spot detection.

        Args:
            min_area_multiplier (float, optional): Increase if some of the small
                shit is detected as a spot. Decrease if some spots are missed.
                Defaults to 0.1.
            max_area_multiplier (float, optional): Increase if some spots are
                missed.  Decrease if some large elements are detected as spots.
                Defaults to 2.
            kernel_size (Tuple[int], optional): Increase with a small downsample
             and vice versa. Defaults to (5, 5).
            max_pixels (int, optional): Downsample the image until the image
                has less than max_pixles pixels. Defaults to 1_000_000.

        Returns:
            Image.Image: TMA spot massk.
        """
        mask = detect_spots(
            image=self._thumbnail,
            mask=self._tissue_mask,
            min_area_multiplier=min_area_multiplier,
            max_area_multiplier=max_area_multiplier,
            kernel_size=kernel_size,
        )
        # Flip for a nicer image.
        mask = 1 - mask
        mask = mask/mask.max()*255
        mask = Image.fromarray(mask.astype(np.uint8))
        return resize(mask, max_pixels)

    def _prepare_directories(self, output_dir: str) -> None:
        out_dir = join(output_dir, self.slide_name)
        # Save paths.
        self._thumb_path = join(out_dir, f'thumbnail_{self.downsample}.jpeg')
        self._annotated_path = join(out_dir, 'thumbnail_annotated.jpeg')
        self._spot_meta_path = join(out_dir, 'spot_metadata.csv')
        self._tile_meta_path = join(out_dir, 'metadata.csv')
        self._image_dir = join(out_dir, 'spots')
        self._tile_dir = join(out_dir, 'tiles')
        self._summary_path = join(out_dir, 'summary.txt')

        # Make dirs.
        os.makedirs(self._image_dir, exist_ok=True)

    def save_spots(
            self,
            output_dir: str,
            overwrite: bool = False,
            image_format: str = 'jpeg',
            quality: int = 95) -> pd.DataFrame:
        """
        Save TMA spots, coordinates and spot numbering.

        Args:
            output_dir (str): Parent directory for all output.
            overwrite (bool, optional): This will **remove** all saved images,
                thumbnail and metadata and save images again.. Defaults to
                False.
            image_format (str, optional): Format can be jpeg or png. Defaults
                to 'jpeg'.
            quality (int, optional): For jpeg compression. Defaults to 95.

        Raises:
            ValueError: Invalid image format.

        Returns:
            pd.DataFrame: Coordinates and spot numbers.
        """
        allowed_formats = ['jpeg', 'png']
        if image_format not in allowed_formats:
            raise ValueError(
                'Image format {} not allowed. Select from {}'.format(
                    image_format, allowed_formats
                ))
        self._prepare_directories(output_dir)
        # Check if slide has been cut before.
        if exists(self._thumb_path) and not overwrite:
            logger.warning(
                'Spots have already been cut! Please set overwrite=True if '
                'you wish to save them again.'
            )
            self.spot_metadata = pd.read_csv(self._spot_meta_path)
            self._spots_saved = True
            return self.spot_metadata
        elif exists(self._thumb_path) and overwrite:
            # Remove all previous files.
            os.remove(self._annotated_path)
            remove_images(self._image_dir)
        # Save text summary.
        with open(self._summary_path, "w") as f:
            f.write(self._summary())
        # Save both thumbnails.
        self._thumbnail.save(self._thumb_path, quality=95)
        self._annotated_thumbnail.save(self._annotated_path, quality=95)
        # Wrap the saving function so it can be parallized.
        func = partial(save_spot, **{
            'slide_path': self.slide_path,
            'image_dir': self._image_dir,
            'image_format': image_format,
            'quality': quality,
        })
        # Multiprocessing to speed things up!
        data = list(zip(self._numbers, self._bounding_boxes))
        spot_paths = []
        with mp.Pool(processes=os.cpu_count()) as p:
            for filepath in tqdm(
                p.imap(func, data),
                total=len(data),
                desc=self.slide_name,
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                spot_paths.append(filepath)
        # Finally save metadata.
        self.spot_metadata['path'] = spot_paths
        self.spot_metadata.to_csv(self._spot_meta_path, index=False)
        self._spots_saved = True
        return self.spot_metadata

    def save_tiles(
        self,
        width: int,
        overlap: float = 0.0,
        max_background: float = 0.999,
        overwrite: bool = False,
        image_format: str = 'jpeg',
        quality: int = 95,
        custom_preprocess: Callable[[Image.Image], dict] = None
    ) -> pd.DataFrame:
        """
        Cut tiles from dearrayed TMA spots.

        Args:
            width (int): Tile width.
            overlap (float, optional): Overlap between neighbouring tiles.
                Defaults to 0.0.
            max_background (float, optional): Maximum amount of background
                allowed for a tile. Defaults to 0.999.
            overwrite (bool, optional): This will **remove** the tiles directory
                completely and save all the tiles again. Defaults to False.
            image_format (str, optional): Format can be jpeg or png. Defaults
                to 'jpeg'.
            quality (int, optional): For jpeg compression. Defaults to 95.
            custom_preprocess (Callable[[Image.Image], dict], optional): This is
                intended for users that want to define their own preprocessing
                function. The function must take a Pillow image as an input and
                return a dictionary of desired metrics. Defaults to None.

        Raises:
            ValueError: Invalid image format.
            IOError: Spots have not been saved first with Dearrayer.save().
            IOError: No spot paths found.

        Returns:
            pd.DataFrame: Metadata.
        """
        allowed_formats = ['jpeg', 'png']
        if image_format not in allowed_formats:
            raise ValueError(
                'Image format {} not allowed. Select from {}'.format(
                    image_format, allowed_formats
                ))
        if not self._spots_saved:
            raise IOError('Please save the spots first with Dearrayer.save()')
        if exists(self._tile_dir) and overwrite == False:
            logger.warning(
                f'{self._tile_dir} already exists! If you want to save tiles '
                'again please set overwrite=True.'
            )
            return pd.read_csv(self._tile_meta_path)
        else:
            # Create the tiles directory
            os.makedirs(self._tile_dir, exist_ok=True)
            # Update summary.txt
            self.width = width
            self.overlap = overlap
            self.max_background = max_background
            with open(self._summary_path, "w") as f:
                f.write(self._summary(cut=True))
        # Let's collect all spot paths.
        spot_paths = self.spot_metadata['path'].tolist()
        # Remove nan paths.
        spot_paths = [x for x in spot_paths if isinstance(x, str)]
        if len(spot_paths) == 0:
            raise IOError('No spot paths found!')
        # Wrap the saving function so it can be parallized.
        func = partial(save_tile, **{
            'image_dir': self._tile_dir,
            'width': width,
            'overlap': overlap,
            'threshold': self.threshold,
            'max_background': max_background,
            'image_format': image_format,
            'quality': quality,
            'custom_preprocess': custom_preprocess
        })
        # Multiprocessing to speed things up!
        metadata = []
        with mp.Pool(processes=os.cpu_count()) as p:
            for results in tqdm(
                p.imap(func, spot_paths),
                total=len(spot_paths),
                desc='Cutting tiles',
                bar_format='{l_bar}{bar:20}{r_bar}{bar:-20b}'
            ):
                metadata.append(results)
        metadata = list(filter(None, metadata))
        metadata = flatten(metadata)
        if len(metadata) == 0:
            logger.warning(f'No tiles saved from any of the spots!')
            return None
        # Save metadata.
        self.tile_metadata = pd.DataFrame(metadata)
        self.tile_metadata['slide_name'] = self.slide_name
        self.tile_metadata.to_csv(self._tile_meta_path, index=False)
        return self.tile_metadata


def save_spot(
        data: Tuple[int, Tuple[int, int, int, int]],
        slide_path: str,
        image_dir: str,
        image_format: str,
        quality: int) -> dict:
    """Saves spot as an image (parallizable)."""
    # Unpack variables
    number, (x, y, w, h) = data
    # Load slide as it can't be pickled...
    reader = OpenSlide(slide_path)
    # Prepare filename.
    filepath = join(image_dir, f'spot-{number}')
    if image_format == 'png':
        filepath = filepath + '.png'
    else:
        filepath = filepath + '.jpeg'
    # Load image.
    try:
        image = reader.read_region((x, y), 0, (w, h)).convert('RGB')
    except:
        logger.warning('Broken slide!')
        return
    # Save image.
    if not exists(filepath):
        image.save(filepath, quality=quality)
    return filepath


def save_tile(
        path: str,
        image_dir: str,
        width: int,
        overlap: float,
        threshold: int,
        max_background: float,
        image_format: str,
        quality: int,
        custom_preprocess: Callable[[Image.Image], dict]) -> dict:
    """Saves tiles from a TMA spot (parallizable)."""
    # Load spot image.
    image = cv2.imread(path)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    # Collect all the tile coordinates.
    coords = _get_all_coordidates(dimensions=image.shape[:2],
                                  width=width, overlap=overlap)
    # Discard based on tissue_mask
    mask = tissue_mask(image=image, threshold=threshold)
    coords = _filter_coordinates(coords, mask, width, max_background)
    if len(coords) == 0:
        return
    # Prepare filename prefix.
    prefix = remove_extension(basename(path))
    # Create spot dir.
    image_dir = join(image_dir, prefix)
    os.makedirs(image_dir, exist_ok=True)
    # Load and save each tile.
    spot_metadata = []
    for (x, y), background in coords:
        # Init empty image.
        tile = np.ones((width, width, 3)) * 255
        # Add data (this way we have padding!).
        tmp = image[y:y+width, x:x+width, :]
        tile[:tmp.shape[0], :tmp.shape[1]] = tmp
        # Turn to PIL image.
        tile = Image.fromarray(tile.astype(np.uint8))
        # Define path.
        tile_path = join(image_dir, f'x-{x}_y-{y}.{image_format}')
        # Collect basic metadata.
        metadata = {
            'path': tile_path,
            'spot_path': path,
            'x': x,
            'y': y,
            'width': width,
            'background': background,
        }
        # Update metadata with preprocessing metrics.
        metadata.update(preprocess(image=tile, threshold=threshold))
        # Add custom metrics.
        if custom_preprocess is not None:
            metadata.update(custom_preprocess(image))
        # Add to spot metadata.
        spot_metadata.append(metadata)
        # Save tile.
        tile.save(tile_path, quality=quality)
    return spot_metadata


def _get_all_coordidates(dimensions, width, overlap):
    """Return all coordinates."""
    x = [0]
    y = [0]
    overlap_px = int(width*overlap)
    while x[-1] < dimensions[0]:
        x.append(x[-1] + width - overlap_px)
    x = x[:-1]
    while y[-1] < dimensions[1]:
        y.append(y[-1] + width - overlap_px)
    y = y[:-1]
    coordinates = list(itertools.product(x, y))
    return coordinates


def _filter_coordinates(coords, mask, width, max_background):
    """Discard coordinates with too much background."""
    filtered = []
    for x, y in coords:
        tile_mask = mask[y:y+width, x:x+width]
        if tile_mask.size == 0:
            continue
        bg_perc = 1 - tile_mask.sum()/tile_mask.size
        if bg_perc < max_background:
            filtered.append(((x, y), bg_perc))
    return filtered
